from __future__ import print_function, division
from numpy import array, float32, isnan, diff, where, add, concatenate, append


def get_gaps_indexes(time, break_tolerance):
    """Array indexes where ``time`` has gaps longer than ``break_tolerance``"""
    gaps = diff(time)
    gaps_indexes = where(gaps > break_tolerance)
    gaps_indexes = add(gaps_indexes, 1) # Off by one :-)
    gaps_indexes = concatenate(gaps_indexes).ravel()  # Flatten
    gaps_indexes = append(array([0]), gaps_indexes)  # Start
    gaps_indexes = append(gaps_indexes, array([len(time)+1]))  # End point
    return gaps_indexes



from __future__ import print_function, division
from numba import jit
import numpy
from numpy import mean, median, full, nan
from wotan.location_estimates import (
    location_iter, hodges, trim_mean, winsorize_mean, hampelfilt, huber_psi, tau
    )
import wotan.constants as constants


@jit(fastmath=True, nopython=True, cache=True)
def running_segment(time, flux, window_length, edge_cutoff, cval, method_code):
    """Iterator for a single time-series segment using time-series window sliders"""

    # Numba can't handle strings, so we're passing the location estimator as an int:
    # 1 : biweight
    # 2 : andrewsinewave
    # 3 : welsch
    # 4 : hodges
    # 5 : median
    # 6 : mean
    # 7 : trim_mean
    # 8 : winsorize
    # 9 : hampelfilt
    # 10: huber_one
    # 11: tau

    size = len(time)
    mean_all = full(size, nan)
    half_window = window_length / 2
    # 0 < Edge cutoff < half_window:
    if edge_cutoff > half_window:
        edge_cutoff = half_window

    # Pre-calculate border checks before entering the loop (reason: large speed gain)
    low_index = numpy.min(time) + edge_cutoff
    hi_index = numpy.max(time) - edge_cutoff
    idx_start = 0
    idx_end = 0

    for i in range(size-1):
        if time[i] > low_index and time[i] < hi_index:
            # Nice style would be:
            #   idx_start = numpy.argmax(time > time[i] - window_length/2)
            #   idx_end = numpy.argmax(time > time[i] + window_length/2)
            # But that's too slow (factor 10). Instead, we write:
            while time[idx_start] < time[i] - half_window:
                idx_start += 1
            while time[idx_end] < time[i] + half_window and idx_end < size-1:
                idx_end += 1
            # Get the location estimate for the segment in question
            # iterative method for: biweight, andrewsinewave, welsch
            f = flux[idx_start:idx_end]
            if method_code == 1 or method_code == 2 or method_code == 3:
                mean_all[i] = location_iter(f, cval, method_code)
            # hodges
            elif method_code == 4:
                mean_all[i] = hodges(f)
            # median
            elif method_code == 5:
                mean_all[i] = median(f)
            # mean
            elif method_code == 6:
                mean_all[i] = mean(f)
            # trim_mean
            elif method_code == 7:
                mean_all[i] = trim_mean(f, proportiontocut=cval)
            # winsorize
            elif method_code == 8:
                mean_all[i] = winsorize_mean(f, proportiontocut=cval)
            # hampel
            elif method_code == 9:
                mean_all[i] = hampelfilt(f, cval=cval)
            # huber_psi
            elif method_code == 10:
                mean_all[i] = huber_psi(f, cval=cval)
            # tau
            elif method_code == 11:
                mean_all[i] = tau(f, cval=cval)
    return mean_all


def running_segment_slow(time, flux, window_length, edge_cutoff, cval, method):
    """With the import, we can't use numba. Thus this here is separate and slower"""
    try:
        # import statsmodels.api as sm
        from statsmodels.api import robust, RLM
    except:
        raise ImportError('Could not import statsmodels')
    size = len(time)
    mean_all = full(size, nan)
    half_window = window_length / 2
    # 0 < Edge cutoff < half_window:
    if edge_cutoff > half_window:
        edge_cutoff = half_window
    # Flat regression for ``hampel```and ``ramsay``
    flat = numpy.ones(len(flux))
    # Pre-calculate border checks before entering the loop (reason: large speed gain)
    low_index = numpy.min(time) + edge_cutoff
    hi_index = numpy.max(time) - edge_cutoff
    idx_start = 0
    idx_end = 0

    for i in range(size-1):
        if time[i] > low_index and time[i] < hi_index:
            # Nice style would be:
            #   idx_start = numpy.argmax(time > time[i] - window_length/2)
            #   idx_end = numpy.argmax(time > time[i] + window_length/2)
            # But that's too slow (factor 10). Instead, we write:
            while time[idx_start] < time[i] - half_window:
                idx_start += 1
            while time[idx_end] < time[i] + half_window and idx_end < size-1:
                idx_end += 1
            # Huber is not numerically stable
            # If it does not converge, we fall back to the median
            if method == 'huber':
                try:
                    huber = robust.scale.Huber(
                        maxiter=constants.MAXITER_HUBER,
                        tol=constants.FTOL,
                        c=cval
                        )
                    mean_all[i], error = huber(flux[idx_start:idx_end])
                except:
                    mean_all[i] = median(flux[idx_start:idx_end])
            elif method == 'hampel':
                points = len(flux[idx_start:idx_end])
                mean_all[i] = RLM(
                    flux[idx_start:idx_end],
                    flat[:points],
                    M=robust.norms.Hampel(
                        a=cval[0],
                        b=cval[1],
                        c=cval[2])
                    ).fit().params[0]
            elif method == 'ramsay': 
                points = len(flux[idx_start:idx_end])
                mean_all[i] = RLM(
                    flux[idx_start:idx_end],
                    flat[:points],
                    M=robust.norms.RamsayE(
                        a=cval)
                    ).fit().params[0]
    return mean_all



from __future__ import print_function, division
import numpy
import scipy.interpolate


def detrend_huber_spline(time, flux, knot_distance):
    """Robust B-Spline regression with scikit-learn"""
    try:
        from sklearn.base import TransformerMixin
        from sklearn.pipeline import make_pipeline
        from sklearn.linear_model import HuberRegressor
    except:
        raise ImportError('Could not import sklearn')

    class BSplineFeatures(TransformerMixin):
        def __init__(self, knots, degree=3, periodic=False):
            self.bsplines = self.get_bspline_basis(
                knots, degree, periodic=periodic
            )
            self.nsplines = len(self.bsplines)

        def fit(self, X, y=None):
            return self

        def transform(self, X):
            nsamples, nfeatures = X.shape
            features = numpy.zeros(
                (nsamples, nfeatures * self.nsplines)
            )
            for ispline, spline in enumerate(self.bsplines):
                istart = ispline * nfeatures
                iend = (ispline + 1) * nfeatures
                features[
                    :, istart:iend
                ] = scipy.interpolate.splev(X, spline)
            return features

        def get_bspline_basis(
            self, knots, degree=3, periodic=False
        ):
            """Get spline coefficients for each basis spline"""
            knots, coeffs, degree = scipy.interpolate.splrep(
                knots,
                numpy.zeros(len(knots)),
                k=degree,
                per=periodic,
            )
            ncoeffs = len(coeffs)
            bsplines = []
            for ispline in range(len(knots)):
                coeffs = [
                    1.0 if ispl == ispline else 0.0
                    for ispl in range(ncoeffs)
                ]
                bsplines.append((knots, coeffs, degree))
            return bsplines

    duration = numpy.max(time) - numpy.min(time)
    no_knots = int(duration / knot_distance)
    knots = numpy.linspace(
        numpy.min(time), numpy.max(time), no_knots
    )

    trend = (
        make_pipeline(
            BSplineFeatures(knots), HuberRegressor()
        )
        .fit(time[:, numpy.newaxis], flux)
        .predict(time[:, None])
    )
    return trend



from numpy import isnan, array, inf


def cleaned_array(t, y, dy=None):
    """Takes numpy arrays with masks and non-float values.
    Returns unmasked cleaned arrays."""

    def isvalid(value):
        valid = False
        if value is not None:
            if not isnan(value):
                if value < inf:
                    valid = True
        return valid

    # Start with empty Python lists and convert to numpy arrays later (reason: speed)
    clean_t = []
    clean_y = []
    if dy is not None:
        clean_dy = []

    # Cleaning numpy arrays with both NaN and None values is not trivial, as the usual
    # mask/delete filters do not accept their simultanous ocurrence without warnings.
    # Instead, we iterate over the array once; this is not Pythonic but works reliably.
    for i in range(len(y)):

        # Case: t, y, dy
        if dy is not None:
            if isvalid(y[i]) and isvalid(t[i]) and isvalid(dy[i]):
                clean_y.append(y[i])
                clean_t.append(t[i])
                clean_dy.append(dy[i])

        # Case: only t, y
        else:
            if isvalid(y[i]) and isvalid(t[i]):
                clean_y.append(y[i])
                clean_t.append(t[i])

    clean_t = array(clean_t, dtype=float)
    clean_y = array(clean_y, dtype=float)

    if dy is None:
        return clean_t, clean_y
    else:
        clean_dy = array(clean_dy, dtype=float)
        return clean_t, clean_y, clean_dy



"""Wotan is a free and open source algorithm to automagically remove stellar trends
from light curves for exoplanet transit detection.
"""

from __future__ import print_function, division
import numpy
from numpy import array, isnan, float32, append, full, where, nan, ones, inf, median
from scipy.signal import savgol_filter, medfilt

# wotan
import wotan.constants as constants
from wotan.cofiam import detrend_cofiam
from wotan.gp import make_gp
from wotan.huber_spline import detrend_huber_spline
from wotan.slider import running_segment, running_segment_slow
from wotan.gaps import get_gaps_indexes
from wotan.t14 import t14
from wotan.pspline import pspline
from wotan.iter_spline import iter_spline


def flatten(time, flux, window_length=None, edge_cutoff=0, break_tolerance=None,
            cval=None, return_trend=False, method='biweight', kernel=None,
            kernel_size=None, kernel_period=None,
            proportiontocut=constants.PROPORTIONTOCUT, robust=False):
    """
    ``flatten`` removes low frequency trends in time-series data.
        
    Parameters
    ----------
    time : array-like
        Time values
    flux : array-like
        Flux values for every time point
    window_length : float
        The length of the filter window in units of ``time`` (usually days), or in
        cadences (for cadence-based sliders ``savgol`` and ``medfilt``).
    method : string, default: ``biweight``
        Detrending method. Rime-windowed sliders: ``median``, ``biweight``, ``hodges``,
        ``tau``, ``welsch``, ``huber``, ``huber_psi``, ``andrewsinewave``, ``mean``, 
        ``hampel``, ``ramsay``, ``trim_mean``, ``hampelfilt``, ``winsorize``. Cadence
        based slider: ``medfilt``. Splines: ``hspline``, ``rspline`, ``pspline``.
        Locally weighted scatterplot smoothing: ``lowess``. Savitzky-Golay filter:
        ``savgol``. Gaussian processes: ``gp``. Cosine Filtering with Autocorrelation 
        Minimization: ``cofiam``.  Friedman's Super-Smoother: ``supersmoother``.
    break_tolerance : float, default: window_length/2
        If there are large gaps in time (larger than ``window_length``/2), flatten will
        split the flux into several sub-lightcurves and apply the filter to each
        individually. ``break_tolerance`` must be in the same unit as ``time`` (usually
        days). To disable this feature, set ``break_tolerance`` to 0. If the method is
        ``supersmoother`` and no ``break_tolerance`` is provided, it will be taken as
        `1` in units of ``time``.
    edge_cutoff : float, default: None
        Trends near edges are less robust. Depending on the data, it may be beneficial
        to remove edges. The ``edge_cutoff`` defines the length (in units of time) to be
        cut off each edge. Default: Zero. Cut off is maximally ``window_length``/2, as
        this fills the window completely. Applicable only for time-windowed sliders.
    cval : float or int
        Tuning parameter for the robust estimators. See documentation for defaults. 
        Larger values for make the estimate more efficient but less robust. For the 
        super-smoother, cval determines the bass enhancement (smoothness) and can be 
        `None` or in the range 0 < ``cval`` < 10. For the ``savgol``, ``cval`` 
        determines the (integer) polynomial order (default: 2).
    proportiontocut : float, default: 0.1
        Fraction to cut off (or filled) of both tails of the distribution using methods
        ``trim_mean`` (or ``winsorize``)
    kernel : str, default: `squared_exp`
        Choice of `squared_exp` (squared exponential), `matern`, `periodic`,
        `periodic_auto`.
    kernel_size : float, default: 1
        The length scale of the Gaussian Process kernel.
    kernel_period : float
        The periodicity of the Gaussian Process kernel (in units of ``time``). Must be
        provided for the kernel `periodic`. Can not be specified for the
        `periodic_auto`, for which it is determined automatically using a Lomb-Scargle
        periodogram pre-search.
    robust : bool, default: False
        If `True`, the Gaussian Process kernels `squared_exp` and `matern` will be run
        iteratively. In each iteration, 2-sigma outliers from the fitted trend will be 
        clipped until convergence.
    return_trend : bool, default: False
        If `True`, the method will return a tuple of two elements
        (``flattened_flux``, ``trend_flux``) where ``trend_flux`` is the removed trend.

    Returns
    -------
    flatten_flux : array-like
        Flattened flux.
    trend_flux : array-like
        Trend in the flux. Only returned if ``return_trend`` is `True`.
    """
    if method not in constants.methods:
        raise ValueError('Unknown detrending method')

    # Numba can't handle strings, so we're passing the location estimator as an int:
    if method == 'biweight':
        method_code = 1
    elif method == 'andrewsinewave':
        method_code = 2
    elif method == 'welsch':
        method_code = 3
    elif method == 'hodges':
        method_code = 4
    elif method == 'median':
        method_code = 5
    elif method == 'mean':
        method_code = 6
    elif method == 'trim_mean':
        method_code = 7
    elif method == 'winsorize':
        method_code = 8
    elif method == 'hampelfilt':
        method_code = 9
    elif method == 'huber_psi':
        method_code = 10
    elif method == 'tau':
        method_code = 11

    error_text = 'proportiontocut must be >0 and <0.5'
    if not isinstance(proportiontocut, float):
        raise ValueError(error_text)
    if proportiontocut >= 0.5 or proportiontocut <= 0:
        raise ValueError(error_text)

    # Default cval values for robust location estimators
    if cval is None:
        if method == 'biweight':
            cval = 5
        elif method == 'andrewsinewave':
            cval = 1.339
        elif method == 'welsch':
            cval = 2.11
        elif method == 'huber':
            cval = 1.5
        elif method == 'huber_psi':
            cval = 1.28
        elif method in ['trim_mean', 'winsorize']:
            cval = proportiontocut
        elif method == 'hampelfilt':
            cval = 3
        elif method == 'tau':
            cval = 4.5
        elif method == 'hampel':
            cval = (1.7, 3.4, 8.5)
        elif method == 'ramsay':
            cval = 0.3
        elif method == 'savgol':  # polyorder
            cval = 2  # int
        else:
            cval = 0  # avoid numba type inference error: None type multi with float

    if cval is not None and method == 'supersmoother':
        if cval > 0 and cval < 10:
            supersmoother_alpha = cval
        else:
            supersmoother_alpha = None

    # Maximum gap in time should be half a window size.
    # Any larger is nonsense,  because then the array has a full window of data
    if window_length is None:
        window_length = 2  # so that break_tolerance = 1 in the supersmoother case
    if break_tolerance is None:
        break_tolerance = window_length / 2
    if break_tolerance == 0:
        break_tolerance = inf

    # Numba is very fast, but doesn't play nicely with NaN values
    # Therefore, we make new time-flux arrays with only the floating point values
    # All calculations are done within these arrays
    # Afterwards, the trend is transplanted into the original arrays (with the NaNs)
    time = array(time, dtype=float32)
    flux = array(flux, dtype=float32)
    mask = isnan(time * flux)
    time_compressed = numpy.ma.compressed(numpy.ma.masked_array(time, mask))
    flux_compressed = numpy.ma.compressed(numpy.ma.masked_array(flux, mask))

    # Get the indexes of the gaps
    gaps_indexes = get_gaps_indexes(time_compressed, break_tolerance=break_tolerance)
    trend_flux = array([])
    trend_segment = array([])

    # Iterate over all segments
    for i in range(len(gaps_indexes) - 1):
        time_view = time_compressed[gaps_indexes[i]:gaps_indexes[i+1]]
        flux_view = flux_compressed[gaps_indexes[i]:gaps_indexes[i+1]]
        methods = ["biweight", "andrewsinewave", "welsch", "hodges", "median", "mean",
            "trim_mean", "winsorize", "huber_psi", "hampelfilt", "tau"]
        if method in methods:
            trend_segment = running_segment(
                time_view,
                flux_view,
                window_length,
                edge_cutoff,
                cval,
                method_code)
        elif method in ["huber", "hampel", "ramsay"]:
            trend_segment = running_segment_slow(
                time_view,
                flux_view,
                window_length,
                edge_cutoff,
                cval,
                method
                )
        elif method == 'lowess':
            try:
                import statsmodels.api
            except:
                raise ImportError('Could not import statsmodels')
            duration = numpy.max(time_compressed) - numpy.min(time_compressed)
            trend_segment = statsmodels.api.nonparametric.lowess(
                endog=flux_view,
                exog=time_view,
                frac=window_length / duration,
                missing='none',
                return_sorted=False
                )
        elif method == 'hspline':
            trend_segment = detrend_huber_spline(
                time_view,
                flux_view,
                knot_distance=window_length)
        elif method == 'supersmoother':
            try:
                from supersmoother import SuperSmoother as supersmoother
            except:
                raise ImportError('Could not import supersmoother')
            win = window_length / (max(time)-min(time))
            trend_segment = supersmoother(
                alpha=supersmoother_alpha,
                primary_spans=(
                    constants.primary_span_lower * win, 
                    win,
                    constants.primary_span_upper * win
                    ),
                middle_span=constants.middle_span * win,
                final_span=constants.upper_span * win
                ).fit(time_view, flux_view,).predict(time_view)
        elif method == 'cofiam':
            trend_segment = detrend_cofiam(
                time_view, flux_view, ones(len(time_view)), window_length)
        elif method == 'savgol':
            if window_length%2 == 0:
                window_length += 1
            trend_segment = savgol_filter(flux_view, window_length, polyorder=int(cval))
        elif method == 'medfilt':
            trend_segment = medfilt(flux_view, window_length)
        elif method == 'gp':
            print('Segment', i + 1, 'of', len(gaps_indexes) - 1)
            trend_segment = make_gp(
                time_view,
                flux_view,
                kernel,
                kernel_size,
                kernel_period,
                robust
                )
        elif method == 'rspline':
            print('Segment', i + 1, 'of', len(gaps_indexes) - 1)
            trend_segment = iter_spline(time_view, flux_view, window_length)
        elif method == 'pspline':
            print('Segment', i + 1, 'of', len(gaps_indexes) - 1)
            trend_segment = pspline(time_view, flux_view)

        trend_flux = append(trend_flux, trend_segment)

    # Insert results of non-NaNs into original data stream
    trend_lc = full(len(time), nan)
    mask = where(~mask)[0]
    for idx in range(len(mask)):
        trend_lc[mask[idx]] = trend_flux[idx]

    flatten_lc = flux / trend_lc
    if return_trend:
        return flatten_lc, trend_lc
    return flatten_lc



from __future__ import print_function, division
import numpy
import wotan.constants as constants
from wotan.helpers import cleaned_array


def pspline(time, flux):
    try:
        from pygam import LinearGAM, s
    except:
        raise ImportError('Could not import pygam')

    newflux = flux.copy()
    newtime = time.copy()
    detrended_flux = flux.copy()

    for i in range(constants.PSPLINES_MAXITER):
        mask_outliers = numpy.ma.where(
            1-detrended_flux < constants.PSPLINES_STDEV_CUT*numpy.std(detrended_flux))
        newtime, newflux = cleaned_array(newtime[mask_outliers], newflux[mask_outliers])
        gam = LinearGAM(s(0, n_splines=constants.PSPLINES_MAX_SPLINES))
        search_gam = gam.gridsearch(newtime[:, numpy.newaxis], newflux, progress=False)
        trend = search_gam.predict(newtime)
        detrended_flux = newflux / trend
        stdev = numpy.std(detrended_flux)
        mask_outliers = numpy.ma.where(
            1-detrended_flux > constants.PSPLINES_STDEV_CUT*numpy.std(detrended_flux))
        print('Iteration:', i + 1, 'Rejected outliers:', len(mask_outliers[0]))

        # Check convergence
        if len(mask_outliers[0]) == 0:
            print('Converged.')
            break

    # Final iteration, applied to unclipped time series (interpolated over clipped values)
    mask_outliers = numpy.ma.where(1-detrended_flux < constants.PSPLINES_STDEV_CUT*stdev)
    newtime, newflux = cleaned_array(newtime[mask_outliers], newflux[mask_outliers])
    gam = LinearGAM(s(0, n_splines=constants.PSPLINES_MAX_SPLINES))
    search_gam = gam.gridsearch(newtime[:, numpy.newaxis], newflux, progress=False)
    trend = search_gam.predict(time)

    return trend



import numpy as np


def clipit(data, low, high, method, center):
    """Clips data in the current segment"""

    # For the center point, take the median (if center_code==0), else the mean
    if center == 'mad':
        mid = np.nanmedian(data)
    else:
        mid = np.nanmean(data)
    data = np.nan_to_num(data)
    diff = data - mid

    if method == 'median':
        cutoff = np.nanmedian(np.abs(data - mid))
    else:
        cutoff = np.nanstd(data)

    # Clip values high (low) times the threshold (in MAD or STDEV)
    data[diff > high * cutoff] = np.nan
    data[diff < -low * cutoff] = np.nan
    return data


def slide_clip(time, data, window_length, low=3, high=3, method=None, center=None):
    """Sliding time-windowed outlier clipper.

    Parameters
    ----------
    time : array-like
        Time values
    flux : array-like
        Flux values for every time point
    window_length : float
        The length of the filter window in units of ``time`` (usually days)
    low : float or int
        Lower bound factor of clipping. Default is 3.
    high : float or int
        Lower bound factor of clipping. Default is 3.
    method : ``mad`` (median absolute deviation; default) or ``std`` (standard deviation)
        Outliers more than ``low`` and ``high`` times the ``mad`` (or the ``std``) from
        the middle point are clipped
    center : ``median`` (default) or ``mean``
        Method to determine the middle point

    Returns
    -------

    clipped : array-like
        Input array with clipped elements removed.
    """

    if method is None:
        method = 'mad'
    if center is None:
        center = 'median'

    low_index = np.min(time)
    hi_index = np.max(time)
    idx_start = 0
    idx_end = 0
    size = len(time)
    half_window = window_length / 2
    clipped_data = np.full(size, np.nan)
    for i in range(size-1):
        if time[i] > low_index and time[i] < hi_index:
            # Nice style would be:
            #   idx_start = numpy.argmax(time > time[i] - window_length/2)
            #   idx_end = numpy.argmax(time > time[i] + window_length/2)
            # But that's too slow (factor 10). Instead, we write:
            while time[idx_start] < time[i] - half_window:
                idx_start += 1
            while time[idx_end] < time[i] + half_window and idx_end < size-1:
                idx_end += 1
            # Clip the current sliding segment
            f = data[idx_start:idx_end]
            clipped_data[idx_start:idx_end] = clipit(f, low, high, method, center)
    return clipped_data


from __future__ import print_function, division

# All supported detrending methods.
# If user requests other method, an error is raised.
methods = [
    "biweight",
    "lowess",
    "andrewsinewave",
    "welsch",
    "hodges",
    "median",
    "mean",
    "trim_mean",
    "hspline",
    "pspline",
    "cofiam",
    "supersmoother",
    "savgol",
    "medfilt",
    "gp",
    "rspline",
    "huber",
    "winsorize",
    "hampelfilt",
    "huber_psi",
    "hampel",
    "tau",
    "ramsay"
]

# astrophysical constants
G = 6.67384e-11  # gravitational constant [m^3 / kg / s^2]
R_sun = 695508000  # radius of the Sun [m]
R_earth = 6371000  # radius of the Earth [m]
R_jup = 69911000  # radius of Jupiter [m]
M_sun = 1.989 * 10 ** 30  # mass of the Sun [kg]
SECONDS_PER_DAY = 86400

# Desired precision of the final location estimate of the `biweight`, `welsch`,
# and `andrewsinewave`. All other methods use one-step estimates. The iterative
# algorithm based on Newton-Raphson stops when the change in location becomes
# smaller than ``FTOL``. Default: `1e-6`, or 1ppm. Higher precision comes at
# greater computational expense.
FTOL = 1e-6

# Iterative Huber estimator sometimes fails to converge. Its default is 30 in:
# https://www.statsmodels.org/dev/_modules/statsmodels/robust/scale.html#Huber
# This is often not sufficient --> set MAXITER=1000 to avoid infinite loop
MAXITER_HUBER = 1000

# Fraction to cut off (or filled) of both tails of the distribution using methods
# ``trim_mean`` (or ``winsorize``)
PROPORTIONTOCUT = 0.1

# Maximum number of sines to be fit for COFIAM
COFIAM_MAX_SINES = 100

# Lomb-Scargle periodogram frequency search grid size
LS_FREQS = 10000


# SuperSmoother span constants
# Similar to 
# https://github.com/jakevdp/supersmoother/blob/master/supersmoother/supersmoother.py
primary_span_lower = 1
primary_span_upper = 3
middle_span = 3
upper_span = 2

# Penalized splines
PSPLINES_MAXITER = 10
PSPLINES_MAX_SPLINES = 50

# Outliers beyond these threshold are removed in each iteration
# The threhold is calculated in standard deviations from the mean (after trend fitting)
PSPLINES_STDEV_CUT = 2  



from __future__ import print_function, division
from numba import jit
import numpy
from numpy import mean, median, inf, sin, exp, pi, array, sort
import wotan.constants as constants


@jit(fastmath=True, nopython=True, cache=True)
def location_iter(data, cval, method_code):
    """Robust location estimators"""

    # Numba can't handle strings, so we're passing the location estimator as an int:
    # 1 : biweight
    # 2 : andrewsinewave
    # 3 : welsch
    # (the others are not relevant for location_iter)

    # Initial estimate for the central location
    delta_center = inf
    median_data = median(data)
    mad = median(numpy.abs(data - median_data))
    center = center_old = median_data

    # Neglecting this case was a bug in scikit-learn
    if mad == 0:
        return center

    # one expensive division here, instead of two per loop later
    cmad = 1 / (cval * mad)

    # Newton-Raphson iteration, where each result is taken as the initial value of the
    # next iteration. Stops when the difference of a round is below ``FTOL`` threshold
    while numpy.abs(delta_center) > constants.FTOL:
        distance = data - center
        dmad = distance * cmad

        # Inlier weights
        # biweight
        if method_code == 1:
            weight = (1 - dmad ** 2) ** 2
        # andrewsinewave
        if method_code == 2:
            # avoid division by zero
            dmad[dmad == 0] = 1e-10
            weight = sin(dmad) / dmad
        # welsch
        if method_code == 3:
            weight = exp(-(dmad**2) / 2)

        # Outliers with weight zero
        # biweight or welsch
        if method_code == 1 or method_code == 3:
            weight[(numpy.abs(dmad) >= 1)] = 0
        # andrewsinewave
        else:
            weight[(numpy.abs(dmad) >= pi)] = 0

        center += numpy.sum(distance * weight) / numpy.sum(weight)

        # Calculate how much center moved to check convergence threshold
        delta_center = center_old - center
        center_old = center
    return center


@jit(fastmath=True, nopython=True, cache=True)
def tau(x, cval):
    """Robust estimators of location and scale, with breakdown points of 50%.
    Also referred to as: Tau measure of location by Yohai and Zamar
    Source: Yohai and Zamar JASA, vol 83 (1988), pp 406-413 and 
            Maronna and Zamar Technometrics, vol 44 (2002), pp. 307-317"""

    med = median(x)
    mad = median(numpy.abs(x - med))
    zscore = 0.675  # Z-score of the 75th percentile of the normal distribution
    s = zscore * mad
    wnom = 0
    wden = 0
    for i in range(len(x)):
        y = (x[i] - med) / s
        temp = (1 - (y / cval)**2)**2
        if abs(temp) <= cval:
            wnom += temp * x[i]
            wden += temp
    return wnom / wden


@jit(fastmath=True, nopython=True, cache=True)
def huber_psi(x, cval):
    """One-step M-estimator of location using Huber's psi"""
    med = median(x)
    mad = median(numpy.abs(x - med))
    y = (x - med) / mad
    y[y > cval] = cval
    y[y < -cval] = -cval
    a = numpy.sum(y)
    b = (numpy.abs(y) <= cval).sum()  # how many values in y < cval?
    return med + mad * a / b


@jit(fastmath=True, nopython=True, cache=True)
def hampelfilt(data, cval):
    """Values beyond (cval * mad {=median absolute deviation}) are replaced with the 
    median. Source: Ronald K. Pearson, Moncef Gabbouj (p. 147), "Nonlinear Digital 
    Filtering with Python: An Introduction" recommend cval=3"""
    
    flux = data.copy()
    median_data = median(flux)
    diff = numpy.abs(flux - median_data)
    mad = median(diff)
    flux[diff >= cval * mad] = median_data

    return mean(flux)


@jit(fastmath=True, nopython=True, cache=True)
def trim_mean(data, proportiontocut):
    """Mean of array after trimming `proportiontocut` from both tails."""
    len_data = len(data)
    sorted_data = sort(data)
    cut_idx = int(len_data * proportiontocut)
    return mean(sorted_data[cut_idx:len_data-cut_idx])


@jit(fastmath=True, nopython=True, cache=True)
def winsorize_mean(data, proportiontocut):
    """Mean of array after winsorizing `proportiontocut` from both tails."""
    sorted_data = sort(data)
    idx = int(proportiontocut * len(data)) + 1
    if idx < 0:
        idx = 0
    sorted_data[:idx] = sorted_data[idx]
    sorted_data[-idx:] = sorted_data[-idx]
    return mean(sorted_data)


@jit(fastmath=True, nopython=True, cache=True)
def hodges(data):
    """Hodges-Lehmann-Sen robust location estimator"""
    i = 0
    j = 0
    len_data = len(data)
    hodges = []
    while i < len_data:
        while j < len_data:
            hodges.append((data[i] + data[j]) / 2)
            j += 1
        i += 1
        j = i
    return median(array(hodges))



from __future__ import print_function, division
from wotan import flatten, t14, slide_clip
import numpy
from astropy.io import fits


def load_file(filename):
    """Loads a TESS *spoc* FITS file and returns TIME, PDCSAP_FLUX"""
    hdu = fits.open(filename)
    time = hdu[1].data['TIME']
    flux = hdu[1].data['PDCSAP_FLUX']
    flux[flux == 0] = numpy.nan
    return time, flux


def main():
    print("Starting tests for wotan...")

    numpy.testing.assert_almost_equal(
        t14(R_s=1, M_s=1, P=365),
        0.6490025258902046)

    numpy.testing.assert_almost_equal(
        t14(R_s=1, M_s=1, P=365, small_planet=True),
        0.5403690143737738)
    print("Transit duration correct.")

    numpy.random.seed(seed=0)  # reproducibility

    print("Slide clipper...")
    points = 1000
    time = numpy.linspace(0, 30, points)
    flux = 1 + numpy.sin(time)  / points
    noise = numpy.random.normal(0, 0.0001, points)
    flux += noise

    for i in range(points):  
        if i % 75 == 0:
            flux[i:i+5] -= 0.0004  # Add some transits
            flux[i+50:i+52] += 0.0002  # and flares

    clipped = slide_clip(
    time,
    flux,
    window_length=0.5,
    low=3,
    high=2,
    method='mad',
    center='median'
    )
    numpy.testing.assert_almost_equal(numpy.nansum(clipped), 948.9926368754939)

    """
    import matplotlib.pyplot as plt
    plt.scatter(time, flux, s=3, color='black')
    plt.scatter(time, clipped, s=3, color='orange')
    plt.show()
    """

    # TESS test
    print('Loading TESS data from archive.stsci.edu...')
    path = 'https://archive.stsci.edu/hlsps/tess-data-alerts/'
    filename = "hlsp_tess-data-alerts_tess_phot_00062483237-s01_tess_v1_lc.fits"
    #path = 'P:/P/Dok/tess_alarm/'
    #filename = "hlsp_tess-data-alerts_tess_phot_00062483237-s01_tess_v1_lc.fits"
    #filename = 'P:/P/Dok/tess_alarm/hlsp_tess-data-alerts_tess_phot_00077031414-s02_tess_v1_lc.fits'
    #filename = 'tess2018206045859-s0001-0000000201248411-111-s_llc.fits'
    time, flux = load_file(path + filename)

    window_length = 0.5
    
    print("Detrending 1 (biweight)...")
    flatten_lc, trend_lc = flatten(
        time,
        flux,
        window_length,
        edge_cutoff=1,
        break_tolerance=0.1,
        return_trend=True,
        cval=5.0)

    numpy.testing.assert_equal(len(trend_lc), 20076)
    numpy.testing.assert_almost_equal(numpy.nanmax(trend_lc), 28754.985299070882)
    numpy.testing.assert_almost_equal(numpy.nanmin(trend_lc), 28615.108124724477)
    numpy.testing.assert_almost_equal(trend_lc[500], 28671.686308143515)

    numpy.testing.assert_equal(len(flatten_lc), 20076)
    numpy.testing.assert_almost_equal(numpy.nanmax(flatten_lc), 1.0034653549250616)
    numpy.testing.assert_almost_equal(numpy.nanmin(flatten_lc), 0.996726610702177)
    numpy.testing.assert_almost_equal(flatten_lc[500], 1.000577429565131)

    print("Detrending 2 (andrewsinewave)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length, method='andrewsinewave', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.15471987987, decimal=2)

    print("Detrending 3 (welsch)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length, method='welsch', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.16764691235, decimal=2)

    print("Detrending 4 (hodges)...")
    flatten_lc, trend_lc = flatten(time[:1000], flux[:1000], window_length, method='hodges', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 994.0110525909206, decimal=2)

    print("Detrending 5 (median)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length, method='median', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.122065014355, decimal=2)

    print("Detrending 6 (mean)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length, method='mean', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.032473037714, decimal=2)

    print("Detrending 7 (trim_mean)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length, method='trim_mean', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.095164910334, decimal=2)

    print("Detrending 8 (supersmoother)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length, method='supersmoother', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18123.00632204841, decimal=2)

    print("Detrending 9 (hspline)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length, method='hspline', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18123.07625225313, decimal=2)

    print("Detrending 10 (cofiam)...")
    flatten_lc, trend_lc = flatten(time[:2000], flux[:2000], window_length, method='cofiam', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 1948.9999999987976, decimal=2)

    print("Detrending 11 (savgol)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length=301, method='savgol', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18123.003465539354, decimal=2)

    print("Detrending 12 (medfilt)...")
    flatten_lc, trend_lc = flatten(time, flux, window_length=301, method='medfilt', return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18123.22609806557, decimal=2)

    print("Detrending 12 (gp squared_exp)...")
    flatten_lc, trend_lc1 = flatten(
        time[:2000],
        flux[:2000],
        method='gp',
        kernel='squared_exp',
        kernel_size=10,
        return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 1948.99958552324, decimal=2)

    print("Detrending 13 (gp squared_exp robust)...")
    flatten_lc, trend_lc1 = flatten(
        time[:2000],
        flux[:2000],
        method='gp',
        kernel='squared_exp',
        kernel_size=10,
        robust=True,
        return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 1948.8820772313468, decimal=2)
    


    print("Detrending 14 (gp matern)...")
    flatten_lc, trend_lc2 = flatten(
        time[:2000],
        flux[:2000],
        method='gp',
        kernel='matern',
        kernel_size=10,
        return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 1949.0001583058202, decimal=2)

    print("Detrending 15 (gp periodic)...")
    flatten_lc, trend_lc2 = flatten(
        time[:2000],
        flux[:2000],
        method='gp',
        kernel='periodic',
        kernel_size=1,
        kernel_period=10,
        return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 1948.9999708985608, decimal=2)

    time_synth = numpy.linspace(0, 30, 200)
    flux_synth = numpy.sin(time_synth) + numpy.random.normal(0, 0.1, 200)
    flux_synth = 1 + flux_synth / 100
    time_synth *= 1.5
    print("Detrending 16 (gp periodic_auto)...")
    flatten_lc, trend_lc2 = flatten(
        time_synth,
        flux_synth,
        method='gp',
        kernel='periodic_auto',
        kernel_size=1,
        return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 200, decimal=1)
    
    print("Detrending 17 (rspline)...")
    flatten_lc, trend_lc2 = flatten(
        time,
        flux,
        method='rspline',
        window_length=1,
        return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18121.812790732245, decimal=2)

    print("Detrending 18 (huber)...")
    flatten_lc, trend_lc = flatten(
        time[:1000],
        flux[:1000],
        method='huber',
        window_length=0.5,
        edge_cutoff=0,
        break_tolerance=0.4,
        return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 994.01102, decimal=2)
    
    print("Detrending 19 (winsorize)...")
    flatten_lc, trend_lc2 = flatten(
        time,
        flux,
        method='winsorize',
        window_length=0.5,
        edge_cutoff=0,
        break_tolerance=0.4,
        proportiontocut=0.1,
        return_trend=True)
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.064587196448, decimal=2)
    
    print("Detrending 20 (pspline)...")
    flatten_lc, trend_lc = flatten(
        time,
        flux,
        method='pspline',
        return_trend=True
        )
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18121.832133916843, decimal=2)
    
    print("Detrending 21 (hampelfilt)...")
    flatten_lc, trend_lc5 = flatten(
        time,
        flux,
        method='hampelfilt',
        window_length=0.5,
        cval=3,
        return_trend=True
        )
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.158072498867, decimal=2)
    
    print("Detrending 22 (lowess)...")
    flatten_lc, trend_lc1 = flatten(
        time,
        flux,
        method='lowess',
        window_length=1,
        return_trend=True
        )
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18123.08085676265, decimal=2)

    print("Detrending 23 (huber_psi)...")
    flatten_lc, trend_lc1 = flatten(
        time,
        flux,
        method='huber_psi',
        window_length=0.5,
        return_trend=True
        )
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.122065014355, decimal=2)

    print("Detrending 24 (tau)...")
    flatten_lc, trend_lc2 = flatten(
        time,
        flux,
        method='tau',
        window_length=0.5,
        return_trend=True
        )
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 18119.02772621119, decimal=2)
    

    import numpy as np
    points = 1000
    time = np.linspace(0, 30, points)
    flux = 1 + np.sin(time)  / points
    noise = np.random.normal(0, 0.0001, points)
    flux += noise

    for i in range(points):  
        if i % 75 == 0:
            flux[i:i+5] -= 0.0004  # Add some transits
            flux[i+50:i+52] += 0.0002  # and flares


    print("Detrending 25a (hampel 17A)...")
    flatten_lc, trend_lc1 = flatten(
        time,
        flux,
        method='hampel',
        cval=(1.7, 3.4, 8.5),
        window_length=0.5,
        return_trend=True
        )

    print("Detrending 25b (hampel 25A)...")
    flatten_lc, trend_lc2 = flatten(
        time,
        flux,
        method='hampel',
        cval=(2.5, 4.5, 9.5),
        window_length=0.5,
        return_trend=True
        )
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 997.9994362858843, decimal=2)

    print("Detrending 26 (ramsay)...")
    flatten_lc, trend_lc3 = flatten(
        time,
        flux,
        method='ramsay',
        cval=0.3,
        window_length=0.5,
        return_trend=True
        )
    numpy.testing.assert_almost_equal(numpy.nansum(flatten_lc), 997.9974021484584, decimal=2)


    """
    import matplotlib.pyplot as plt
    plt.scatter(time, flux, s=1, color='black')
    plt.plot(time[:len(trend_lc1)], trend_lc1, color='blue', linewidth=2)
    plt.plot(time[:len(trend_lc1)], trend_lc2, color='red', linewidth=2, linestyle='dashed')
    plt.show()
    plt.close()
    #plt.scatter(time, flatten_lc, s=1, color='black')
    #plt.show()
    """


    print('All tests completed.')


if __name__ == '__main__':
    main()



from __future__ import print_function, division
import numpy
import wotan.constants as constants
from scipy.signal import lombscargle
from wotan.helpers import cleaned_array


def make_gp(time, flux, kernel, kernel_size, kernel_period, robust):
    try:
        from sklearn.gaussian_process import GaussianProcessRegressor
        from sklearn.gaussian_process.kernels import RBF, Matern, ExpSineSquared
    except:
        raise ImportError('Could not import sklearn')

    if kernel_size is None:
        raise ValueError('kernel_size must be specified')
    if not isinstance(kernel_size, float) and not isinstance(kernel_size, int):
        raise ValueError('kernel_size must be a floating point value')
    if kernel_size <= 0 or kernel_size >= float("inf"):
        raise ValueError('kernel_size must be finite and positive')

    # GPs need flux near zero, otherwise they often fail to converge
    # So, we normalize by some constant (the median) and later transform back
    offset = numpy.median(flux)
    flux -= offset

    # RBF and matern kernels are similar when matern's (kernel_size * 1000)
    if kernel == 'matern':
        kernel_size *= 1000

    kernel_size_bounds = (0.5 * kernel_size, 2 * kernel_size)
    grid = time.reshape(-1, 1)

    if kernel is None or kernel == 'squared_exp':
        use_kernel = RBF(kernel_size, kernel_size_bounds)

    if kernel == 'matern':
        use_kernel = Matern(kernel_size, kernel_size_bounds, nu=3/2)

    # Single pass or iteratively clipped
    if kernel == 'matern' or kernel == 'squared_exp' or kernel is None:
        if robust:
            converged = False
            newflux = flux.copy()
            newtime = time.copy()
            detrended_flux = flux.copy()
            for i in range(constants.PSPLINES_MAXITER):
                # Flux must be ~1. First round may by ~0. Then correct: 
                if abs(numpy.median(detrended_flux)) < 0.5:
                    detrended_flux += 1
                mask_outliers = numpy.ma.where(
                    1-detrended_flux < constants.PSPLINES_STDEV_CUT * numpy.std(
                        detrended_flux))
                newtime, newflux = cleaned_array(
                    newtime[mask_outliers], newflux[mask_outliers])
                grid = newtime.reshape(-1, 1)
                GP = GaussianProcessRegressor(use_kernel).fit(grid, newflux)
                trend_segment = GP.predict(grid)
                detrended_flux = (newflux + offset) / (trend_segment + offset)
                mask_outliers = numpy.ma.where(
                    1-detrended_flux > constants.PSPLINES_STDEV_CUT * numpy.std(
                        detrended_flux))
                print('Iteration:', i + 1, 'Rejected outliers:', len(mask_outliers[0]))
                # Check convergence
                if converged or len(mask_outliers[0]) == 0:
                    print('Converged.')
                    break
            # Final iteration, applied to unclipped time series 
            # (interpolated over clipped values)
            trend_segment = GP.predict(time.reshape(-1, 1))
        else:
            trend_segment = GaussianProcessRegressor(
                use_kernel).fit(grid, flux).predict(grid)

    # Single pass; currently no iterative periodic kernel implemented
    if 'periodic' in kernel:
        # Determine most significant period
        if kernel == 'periodic_auto':
            time_span = numpy.max(time) - numpy.min(time)
            cadence = numpy.nanmedian(numpy.diff(time))
            freqs = numpy.geomspace(1/time_span, 1/cadence, constants.LS_FREQS)
            pgram = lombscargle(time, flux, freqs)
            kernel_period = 1 / freqs[numpy.argmax(pgram)] * 2 * numpy.pi
        else:
            if kernel_period is None:
                raise ValueError('kernel_period must be specified')
            if not isinstance(kernel_period, float) and not isinstance(kernel_period, int):
                raise ValueError('kernel_period must be a floating point value')
            if kernel_period <= 0 or kernel_period >= float("inf"):
                raise ValueError('kernel_period must be finite and positive')
        kernel_period_bounds=(0.5 * kernel_period, 2 * kernel_period)
        # The periodic part
        use_kernel = ExpSineSquared(
            kernel_size,
            kernel_period,
            kernel_size_bounds,
            kernel_period_bounds
            )
        # For additional trends
        use_kernel2 = RBF(kernel_size, kernel_size_bounds)  
        trend_segment = GaussianProcessRegressor(
            use_kernel + use_kernel2).fit(grid, flux).predict(grid)

    return (trend_segment + offset)



from __future__ import print_function, division
from numba import jit
from scipy.optimize import leastsq
import numpy
from numpy import (
    mean,
    array,
    inf,
    sin,
    cos,
    pi,
    ones,
    zeros,
)
import wotan.constants as constants


@jit(fastmath=True, nopython=True, cache=True)
def cofiam_detrend_curve(t, fact, D_max, k_max):
    ph = 2 * pi * t / D_max
    detr = ones(len(t)) * fact[0]
    for k in range(1, k_max + 1):
        detr += fact[2 * k - 1] * sin(ph * k) + fact[
            2 * k
        ] * cos(ph * k)
    return detr


def find_detrending_for_region(t, y, ferr, k_max, D_max):
    def fit_func(fact):
        return (
            y - cofiam_detrend_curve(t, fact, D_max, k_max)
        ) / ferr

    def out_func(blub):
        return cofiam_detrend_curve(
            t, best_param, D_max, k_max
        )

    x0 = zeros(2 * k_max + 1)
    x0[0] = mean(y)
    best_param, param_fit_status = leastsq(
        fit_func, x0, ftol=constants.FTOL
    )
    result = cofiam_detrend_curve(
        t, best_param, D_max, k_max
    )
    return out_func


def detrend_cofiam(t, y, ferr, window_length):
    D_max = 2 * (max(t) - min(t))
    k_max = max(
        1,
        min(
            constants.COFIAM_MAX_SINES,
            int(D_max / window_length),
        ),
    )
    dw_previous = inf
    dw_mask = array([True] * len(t))
    for k_m in range(1, k_max + 1):
        detrend_func_temp = find_detrending_for_region(
            t, y, ferr, k_m, D_max
        )
        trend = detrend_func_temp(t)

        # Durbin-Watson autocorrelation statistics
        dw_y = y[dw_mask] / trend[dw_mask] - 1
        dw = numpy.abs(
            numpy.sum((dw_y[1:] - dw_y[:-1]) ** 2)
            / (numpy.sum(dw_y ** 2))
            - 2
        )

        # If Durbin-Watson *increased* this round: Previous was the best
        if dw > dw_previous:
            return trend
        detrend_func = detrend_func_temp
        dw_previous = dw
    return detrend_func(t)



from setuptools import setup
from os import path

# Pull wotan version from single source of truth file
try:  # Python 2
    execfile(path.join("wotan", 'version.py'))
except:  # Python 3
    exec(open(path.join("wotan", 'version.py')).read())

# If Python3: Add "README.md" to setup. 
# Useful for PyPI (pip install wotan). Irrelevant for users using Python2
try:
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except:
    long_description = ' '

setup(
    name='wotan',
    version=WOTAN_VERSIONING,
    description='Detrending algorithms',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/hippke/wotan',
    author='Michael Hippke',
    author_email='michael@hippke.org',
    license='MIT',
    zip_safe=False,
    packages=['wotan'],
    install_requires=[
        'numpy',
        'numba',
        'scipy'
        ]
)


from __future__ import print_function, division
import numpy
import wotan.constants as constants
from wotan.helpers import cleaned_array
from scipy import interpolate


def iter_spline(time, flux, window_length):
    no_knots = (max(time) - min(time)) / window_length
    newflux = flux.copy()
    newtime = time.copy()
    detrended_flux = flux.copy()
    for i in range(constants.PSPLINES_MAXITER):
        mask_outliers = numpy.ma.where(
            1 - detrended_flux < constants.PSPLINES_STDEV_CUT * numpy.std(detrended_flux))
        newtime, newflux = cleaned_array(newtime[mask_outliers], newflux[mask_outliers])
        # knots must not be at the edges, so we take them as [1:-1]
        knots = numpy.linspace(min(newtime), max(newtime), no_knots)[1:-1]
        s = interpolate.LSQUnivariateSpline(newtime, newflux, knots)
        trend_segment = s(newtime)
        detrended_flux = newflux / trend_segment
        mask_outliers = numpy.ma.where(
            1 - detrended_flux > constants.PSPLINES_STDEV_CUT * numpy.std(detrended_flux))
        print('Iteration:', i + 1, 'Rejected outliers:', len(mask_outliers[0]))
        # Check convergence
        if len(mask_outliers[0]) == 0:
            print('Converged.')
            break
    # Final iteration applied all data interpolated over clipped values
    trend_segment = s(time)
    return trend_segment



from __future__ import print_function, division
import numpy
import wotan.constants as constants



def t14(R_s, M_s, P, small_planet=False):
    """Planetary transit duration assuming a central transit on a circular orbit
    
    Parameters
    ----------
    R_s : float
        Stellar radius (in solar radii)
    M_s : float
        Stellar mass (in solar masses)
    P : float
        Planetary period (in days)
    small_planet : bool, default: `False`
        Small planet assumption if `True` (zero planetary radius). Otherwise, a 
        planetary radius of 2 Jupiter radii is assumed (maximum reasonable planet size)
    
    Returns
    -------
    t14 : float
        Planetary transit duration (in days)
    """
    if small_planet:
        planet_size = 0
    else:
        planet_size = 2 * constants.R_jup
    t14 = (
        (constants.R_sun * R_s + planet_size)
        * (
            (4 * P * constants.SECONDS_PER_DAY)
            / (numpy.pi * constants.G * constants.M_sun * M_s)
        )
        ** (1 / 3)
    ) / constants.SECONDS_PER_DAY
    return t14
